import base64
import json
import os
import subprocess
import sys

import yaml

from charmhelpers.core import hookenv, host
from charmhelpers.core.templating import render
from charmhelpers.core.services.base import ManagerCallback
from charmhelpers.fetch import apt_install, apt_update, add_source

from jinja2 import Template


CONFIG_FILE = '/etc/telegraf/telegraf.conf'

CONFIG_DIR = '/etc/telegraf/telegraf.d'


class Restart(ManagerCallback):
    """Restart the service, only if the config changed"""

    def _update_persisted_data(self, key, value):
        """Persist the given 'value' for the given 'key' and return the old value.

        @param key: The key to update.
        @param value: The value to persist.
        @return: The old value of the key, or None if it's a new value.
        """
        config = hookenv.config()
        old = config.get(key, None)
        config[key] = value
        config.save()
        return old

    def __call__(self, manager, service_name, event_name):
        if hookenv.hook_name() in ('update-status',):
            return
        new_hash = host.file_hash(CONFIG_FILE)
        old_hash = self._update_persisted_data('config_hash', new_hash)
        if new_hash != old_hash:
            hookenv.log('Restarting service, config file changed')
            host.service_restart(service_name)
            return
        types = hookenv.relation_types()
        for rel_type in types:
            if rel_type == 'juju-info':
                continue
            config_path = os.path.join('{}/{}.conf'.format(CONFIG_DIR, rel_type))
            new_hash = host.file_hash(config_path)
            old_hash = self._update_persisted_data(rel_type, new_hash)
            if new_hash != old_hash:
                hookenv.log('Restarting service, {}'.format(hookenv.hook_name()))
                host.service_restart(service_name)
                return
        hookenv.log('Not restarting service, config and relations unchanged')


def install():
    hookenv.log('Installing telegraf')
    config = hookenv.config()
    if config['package_name']:
        pkg_file = os.path.join(hookenv.charm_dir(), "files",
                                config['package_name'])
        if os.path.exists(pkg_file):
            env = os.environ.copy()
            env['DEBIAN_FRONTEND'] = 'noninteractive'
            if subprocess.call(["dpkg", "-i", "--force-confold", pkg_file], env=env) != 0:
                hookenv.log("CHARM: Error installing telegraf package")
                sys.exit(1)
        else:
            if config['apt_repository'] and config['apt_repository_key']:
                add_source(config['apt_repository'],
                           config['apt_repository_key'])
                apt_update()
            apt_install(config['package_name'], options=['--force-yes'])


def update_telegraf(service_name):
    config = hookenv.config()
    if hookenv.hook_name() == 'config-changed' \
            and config.changed('package_name'):
        install()
    elif hookenv.hook_name() == 'upgrade-charm':
        install()


def log_start(service_name):
    hookenv.log('telegraf starting')


def render_config(service_name):
    config = hookenv.config()
    extra_options = get_extra_options()
    context = config.copy()
    inputs = base64.b64decode(config['inputs_config'])
    outputs = base64.b64decode(config['outputs_config'])
    tags = []
    if config['tags']:
        for tag in config['tags'].split(','):
            key, value = tag.split("=")
            tags.append('{} = "{}"'.format(key, value))
    context["tags"] = tags
    if inputs:
        context["inputs"] = inputs
    else:
        # use base inputs from charm templates
        with open('templates/base_inputs.conf', 'r') as fd:
            context["inputs"] = render_template(
                fd.read(),
                {'extra_options': extra_options['inputs']})
    if outputs:
        context["outputs"] = outputs
    else:
        context["outputs"] = ""
        hookenv.log("No output plugins in main config.")
    remote_unit_name = get_remote_unit_name()
    if config["hostname"] == "UNIT_NAME":
        if remote_unit_name is not None:
            context["hostname"] = remote_unit_name.replace('/', '-')
        else:
            hookenv.log("Waiting for relation to render config file.")
            # if UNIT_NAME in hostname config and relation not yet available,
            # make telegraf unable to start to not get weird metrics names
            if os.path.exists(CONFIG_FILE):
                os.unlink(CONFIG_FILE)
            return
    hookenv.log("Updating config file")
    render(source='telegraf.conf.tmpl', target=CONFIG_FILE, context=context)


def get_remote_unit_name():
    types = hookenv.relation_types()
    for rel_type in types:
        rels = hookenv.relations_of_type(rel_type)
        if rels and len(rels) >= 1:
            return rels[0]['__unit__']


def get_extra_options():
    extra_options = {'inputs': {}, 'outpus': {}}
    extra_options_raw = hookenv.config()['extra_options']
    extra_opts = yaml.load(extra_options_raw) or {}
    extra_options.update(extra_opts)
    return extra_options


def render_extra_options(kind, name):
    template = """
  {% if extra_options %}
  {% for key, value in extra_options.items() %}
  {% if key == 'tagpass' or key == 'tagdrop' %}
  [{{ kind }}.{{ name }}.{{key}}]
    {% for tag, tagvalue in value.items() %}
    {{ tag }} = {{ tagvalue }}
    {% endfor %}
  {% else %}
  {{ key }} = {{ value }}
  {% endif %}
  {% endfor %}
  {% endif %}
  """
    extra_options = get_extra_options()
    context = {"extra_options": extra_options[kind].get(name, {}),
               "kind": kind,
               "name": name}
    return render_template(template, context)


def render_template(template, context):
    tmpl = Template(template, lstrip_blocks=True, trim_blocks=True)
    return tmpl.render(**context)


def elasticsearch_input(service_name):
    template = """
[[inputs.elasticsearch]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type('elasticsearch')
    hosts = []
    inputs = []
    for rel in rels:
        es_host = rel.get('host')
        port = rel.get('port')
        if not es_host or not port:
            hookenv.log('No host received for relation: {}.'.format(rel))
            continue
        hosts.append("http://{}:{}".format(es_host, port))
    config_path = '{}/{}.conf'.format(CONFIG_DIR, 'elasticsearch')
    if hosts:
        context = {"servers": json.dumps(hosts)}
        inputs.append(render_template(template, context) + \
                      render_extra_options("inputs", "elasticsearch"))
        hookenv.log("Updating {} plugin config file".format('elasticsearch'))
        host.write_file(config_path, '\n'.join(inputs))
    elif os.path.exists(config_path):
        os.unlink(config_path)


def mongodb_input(service_name):
    template = """
[[inputs.mongodb]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type('mongodb')
    mongo_addresses = []
    inputs = []
    for rel in rels:
        addr = rel['private-address']
        port = rel.get('port', None)
        if port:
            mongo_address = '{}:{}'.format(addr, port)
        else:
            mongo_address = addr
        mongo_addresses.append(mongo_address)
    config_path = '{}/{}.conf'.format(CONFIG_DIR, 'mongodb')
    if mongo_addresses:
        extra_options = get_extra_options()
        context = {"servers": json.dumps(mongo_addresses)}
        inputs.append(render_template(template, context) + \
                      render_extra_options("inputs", "mongodb"))
        hookenv.log("Updating {} plugin config file".format('mongodb'))
        host.write_file(config_path, '\n'.join(inputs))
    elif os.path.exists(config_path):
        os.unlink(config_path)


def postgresql_input(service_name):
    template = """
[[inputs.postgresql]]
  address = "host={{host}} user={{user}} password={{password}} dbname={{database}}"
"""
    required_keys = ['host', 'user', 'password', 'database']
    rels = hookenv.relations_of_type('postgresql')
    extra_options = get_extra_options()
    inputs = []
    for rel in rels:
        if all([rel.get(key) for key in required_keys]) \
                and hookenv.local_unit() in rel.get('allowed-units'):
            context = rel.copy()
            inputs.append(render_template(template, context) + \
                          render_extra_options("inputs", "postgresql"))
    config_path = '{}/{}.conf'.format(CONFIG_DIR, 'postgresql')
    if inputs:
        hookenv.log("Updating {} plugin config file".format('postgresql'))
        host.write_file(config_path, '\n'.join(inputs))
    elif os.path.exists(config_path):
        os.unlink(config_path)


def haproxy_input(service_name):
    template = """
[[inputs.haproxy]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type('haproxy')
    haproxy_addresses = []
    inputs = []
    for rel in rels:
        enabled = rel.get('enabled', False)
        if isinstance(enabled, basestring):
            if enabled in ['y', 'yes', 'true', 't', 'on', 'True']:
                enabled = True
            else:
                enabled = False
        if not enabled:
            continue
        addr = rel['private-address']
        if addr == hookenv.unit_private_ip():
            addr = "localhost"
        port = rel['port']
        user = rel['user']
        password = rel.get('password', None)
        userpass = user
        if password:
            userpass += ":{}".format(password)
        haproxy_address = 'http://{}@{}:{}'.format(userpass, addr, port)
        haproxy_addresses.append(haproxy_address)
    config_path = '{}/{}.conf'.format(CONFIG_DIR, 'haproxy')
    if haproxy_addresses:
        inputs.append(render_template(template, {"servers": json.dumps(haproxy_addresses)}) + \
                      render_extra_options("inputs", "haproxy"))
        hookenv.log("Updating {} plugin config file".format('haproxy'))
        host.write_file(config_path, '\n'.join(inputs))
    elif os.path.exists(config_path):
        os.unlink(config_path)


def influxdb_api_output(service_name):
    required_keys = ['hostname', 'port', 'user', 'password']
    rels = hookenv.relations_of_type('influxdb-api')
    endpoints = []
    user = None
    password = None
    for rel in rels:
        if all([rel.get(key) for key in required_keys]):
            endpoints.append("http://{}:{}".format(rel['hostname'], rel['port']))
            if user is None:
                user = rel['user']
            if password is None:
                password = rel['password']
    config_path = '{}/{}.conf'.format(CONFIG_DIR, 'influxdb-api')
    if endpoints:
        hookenv.log("Updating {} plugin config file".format('influxdb-api'))
        content = render(source='influxdb-api.conf.tmpl', target=None,
                         context={'urls': json.dumps(endpoints),
                                  'username': '{}'.format(user),
                                  'password': '{}'.format(password)})
        extra_opts = render_extra_options("outputs", "influxdb")
        host.write_file(config_path, '\n'.join([content, extra_opts]))
    elif os.path.exists(config_path):
        os.unlink(config_path)

